<?php

namespace Drupal\k8s\Form;

use Drupal\cloud\Plugin\cloud\config\CloudConfigPluginManagerInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Form\FormBase;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Messenger\MessengerInterface;
use Drupal\Core\Routing\RouteMatchInterface;
use Drupal\Core\Session\AccountInterface;
use Drupal\k8s\Service\K8sServiceException;
use Drupal\k8s\Service\K8sServiceInterface;
use Drupal\user\Entity\Role;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Form controller for creating project form.
 *
 * @ingroup k8s
 */
class K8sCreateProjectForm extends FormBase {

  /**
   * The K8s Service.
   *
   * @var \Drupal\k8s\Service\K8sServiceInterface
   */
  protected $k8sService;

  /**
   * The messenger Service.
   *
   * @var \Drupal\Core\Messenger\MessengerInterface
   */
  protected $messenger;

  /**
   * The Entity Type Manager.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected $entityTypeManager;

  /**
   * The cloud service provider plugin manager (CloudConfigPluginManager).
   *
   * @var \Drupal\cloud\Plugin\cloud\config\CloudConfigPluginManagerInterface
   */
  protected $cloudConfigPluginManager;

  /**
   * The current user.
   *
   * @var \Drupal\Core\Session\AccountInterface
   */
  protected $currentUser;

  /**
   * The route match.
   *
   * @var \Drupal\Core\Routing\RouteMatchInterface
   */
  protected $routeMatch;

  /**
   * K8sCloudContentForm constructor.
   *
   * @param \Drupal\k8s\Service\K8sServiceInterface $k8s_service
   *   The K8s Service.
   * @param \Drupal\Core\Messenger\MessengerInterface $messenger
   *   The Messenger Service.
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entity_type_manager
   *   The Entity Type Manager.
   * @param \Drupal\cloud\Plugin\cloud\config\CloudConfigPluginManagerInterface $cloud_config_plugin_manager
   *   The cloud service provider plugin manager (CloudConfigPluginManager).
   * @param \Drupal\Core\Session\AccountInterface $current_user
   *   The current user.
   * @param \Drupal\Core\Routing\RouteMatchInterface $route_match
   *   The route match.
   */
  public function __construct(
    K8sServiceInterface $k8s_service,
    MessengerInterface $messenger,
    EntityTypeManagerInterface $entity_type_manager,
    CloudConfigPluginManagerInterface $cloud_config_plugin_manager,
    AccountInterface $current_user,
    RouteMatchInterface $route_match
  ) {
    $this->k8sService = $k8s_service;
    $this->messenger = $messenger;
    $this->entityTypeManager = $entity_type_manager;
    $this->cloudConfigPluginManager = $cloud_config_plugin_manager;
    $this->currentUser = $current_user;
    $this->routeMatch = $route_match;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container) {
    return new static(
      $container->get('k8s'),
      $container->get('messenger'),
      $container->get('entity_type.manager'),
      $container->get('plugin.manager.cloud_config_plugin'),
      $container->get('current_user'),
      $container->get('current_route_match')
    );
  }

  /**
   * {@inheritdoc}
   */
  public function getFormId() {
    return 'k8s_import_project_form';
  }

  /**
   * {@inheritdoc}
   */
  public function buildForm(array $form, FormStateInterface $form_state, $cloud_context = '') {

    $form['project'] = [
      '#type' => 'details',
      '#title' => $this->t('Project'),
      '#description'    => $this->t('Create a project associated with a namespace and role.'),
      '#open' => TRUE,
    ];

    $form['project']['project_name'] = [
      '#type'           => 'textfield',
      '#title'          => $this->t('Project Name'),
      '#description'    => $this->t("The project name must consist of lower case alphanumeric characters or '-', and must start and end with an alphanumeric character (e.g. 'my-name', or '123-abc', regex used for validation is '[a-z0-9]([-a-z0-9]*[a-z0-9])?')"),
      '#maxlength'      => 255,
      '#size'           => 60,
      '#default_value'  => $this->getDefaultProjectName($cloud_context),
      '#required'       => TRUE,
    ];

    $form['project']['username'] = [
      '#type'           => 'textfield',
      '#title'          => $this->t('Username'),
      '#maxlength'      => 255,
      '#size'           => 60,
      '#default_value'  => $this->currentUser->getAccountName(),
      '#required'       => TRUE,
    ];

    $form['k8s_clusters'] = [
      '#type' => 'details',
      '#title' => $this->t('K8s Clusters'),
      '#open' => TRUE,
    ];

    $k8s_cloud_configs = $this->cloudConfigPluginManager->loadConfigEntities('k8s');
    $k8s_clusters_options = [];
    foreach ($k8s_cloud_configs as $k8s_cloud_config) {
      $k8s_clusters_options[$k8s_cloud_config->getCloudContext()] = $k8s_cloud_config->getName();
    }
    $form['k8s_clusters']['k8s_clusters'] = [
      '#type' => 'checkboxes',
      '#options' => $k8s_clusters_options,
      '#default_value'  => array_keys($k8s_clusters_options),
      '#required' => TRUE,
    ];

    $form['actions']['#type'] = 'actions';
    $form['actions']['submit'] = [
      '#type'           => 'submit',
      '#value'          => $this->t('Create'),
    ];

    return $form;
  }

  /**
   * {@inheritdoc}
   */
  public function validateForm(array &$form, FormStateInterface $form_state) {
    $cloud_context = $this->routeMatch->getParameter('cloud_context');
    $this->k8sService->setCloudContext($cloud_context);
    $project_name = $form_state->getValue('project_name');

    // Validate project name.
    $role = Role::load($project_name);
    if (!empty($role)) {
      $form_state->setErrorByName(
        'project_name',
        $this->t('The role "@name" already exists with the same project name.', [
          '@name' => $project_name,
        ])
      );
      return;
    }

    $projects = $this->entityTypeManager->getStorage('k8s_namespace')
      ->loadByProperties([
        'cloud_context' => $cloud_context,
        'name' => $project_name,
      ]);
    if (!empty($projects)) {
      $form_state->setErrorByName(
        'project_name',
        $this->t('The k8s namespace "@name" already exists with the same project name.', [
          '@name' => $project_name,
        ])
      );
    }
  }

  /**
   * {@inheritdoc}
   */
  public function submitForm(array &$form, FormStateInterface $form_state) {
    $cloud_context = $this->routeMatch->getParameter('cloud_context');
    $this->k8sService->setCloudContext($cloud_context);
    $project_name = $form_state->getValue('project_name');
    $k8s_clusters = $form_state->getValue('k8s_clusters');

    try {
      // Create K8s namespace.
      $params = [];
      $params['metadata']['name'] = $project_name;
      $this->k8sService->createNamespace($params);
      $this->k8sService->updateNamespaces([
        'metadata.name' => $project_name,
      ], FALSE);

      // Create Drupal role.
      $data = [
        'id' => $project_name,
        'label' => $project_name,
      ];
      $role = Role::create($data);
      $role->save();

      // Grant permissions.
      $this->grantPermissions($role, $k8s_clusters, $project_name);

      $message = $this->t('The project "%name" has been created.', [
        '%name' => $project_name,
      ]);
      $this->messenger->addMessage($message);

      $form_state->setRedirect('view.k8s_namespace.list', ['cloud_context' => $cloud_context]);
    }
    catch (K8sServiceException $e) {
      $message = $this->t('The project "%name" couldn\'t be created.', [
        '%name' => $project_name,
      ]);
      $this->messenger->addError($message);
    }
  }

  /**
   * Grant permissions.
   *
   * @param \Drupal\user\Entity\Role $role
   *   The role entity.
   * @param array $k8s_clusters
   *   The k8s clusters.
   * @param string $namespace
   *   The namespace.
   */
  private function grantPermissions(Role $role, array $k8s_clusters, $namespace) {
    $permissions = [];

    // Dashboard.
    $permissions[] = 'access dashboard';

    // K8s clusters.
    foreach ($k8s_clusters as $k8s_cluster) {
      $permissions[] = "view $k8s_cluster";
    }

    // Server template.
    $permissions[] = 'list cloud server template';
    $permissions[] = 'add cloud server templates';
    $permissions[] = 'delete own cloud server templates';
    $permissions[] = 'edit own cloud server templates';
    $permissions[] = 'view own published cloud server templates';
    $permissions[] = 'view own unpublished cloud server templates';
    $permissions[] = 'access cloud server template revisions';
    $permissions[] = 'revert all cloud server template revisions';
    $permissions[] = 'delete all cloud server template revisions';
    $permissions[] = 'launch cloud server template';

    // Access namespace.
    $permissions[] = "view k8s namespace $namespace";

    // Resources CRUD.
    $resources = [
      'namespace',
      'pod',
      'deployment',
      'replica set',
      'cron job',
      'job',
      'service',
      'config map',
      'secret',
      'stateful set',
      'ingress',
      'daemon set',
      'endpoint',
      'persistent volume claim',
      'cluster role binding',
      'api service',
    ];
    foreach ($resources as $resource) {
      $permissions[] = "add k8s $resource";
      $permissions[] = "list k8s $resource";

      if ($resource === 'pod') {
        $permissions[] = "view any k8s $resource";
        $permissions[] = "view own k8s $resource";

        $permissions[] = "edit any k8s $resource";
        $permissions[] = "edit own k8s $resource";

        $permissions[] = "delete any k8s $resource";
        $permissions[] = "delete own k8s $resource";
      }
      else {
        $permissions[] = "view k8s $resource";
        $permissions[] = "edit k8s $resource";
        $permissions[] = "delete k8s $resource";
      }
    }

    foreach ($permissions as $permission) {
      $role->grantPermission($permission);
    }

    $role->save();
  }

  /**
   * Get default project name.
   *
   * @param string $cloud_context
   *   The cloud context.
   *
   * @return string
   *   The default project name.
   */
  private function getDefaultProjectName($cloud_context) {
    $namespaces = $this->entityTypeManager->getStorage('k8s_namespace')
      ->loadByProperties([
        'cloud_context' => $cloud_context,
      ]);

    $names = array_map(static function ($namespace) {
      return $namespace->getName();
    }, $namespaces);

    $username = $this->currentUser->getAccountName();

    $number = 1;
    while (in_array($project_name = "$username-project-$number", $names)) {
      $number++;
    }

    return $project_name;
  }

}
