<?php

namespace Drupal\cloud\Form\Config;

use Drupal\Core\Config\ConfigFactoryInterface;
use Drupal\Core\Extension\ModuleHandlerInterface;
use Drupal\Core\Form\ConfigFormBase;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Routing\UrlGeneratorInterface;
use Drupal\geocoder\ProviderPluginManager;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Class Cloud Admin Settings.
 */
class CloudAdminSettings extends ConfigFormBase {

  /**
   * The module handler.
   *
   * @var \Drupal\Core\Extension\ModuleHandlerInterface
   */
  protected $moduleHandler;

  /**
   * The geocoder provider plugin manager.
   *
   * @var \Drupal\geocoder\ProviderPluginManager
   */
  protected $geocoderProviderPluginManager;

  /**
   * The url generator.
   *
   * @var \Drupal\Core\Routing\UrlGeneratorInterface
   */
  protected $urlGenerator;

  /**
   * Constructs a CloudAdminSettings instance.
   *
   * @param \Drupal\Core\Config\ConfigFactoryInterface $config_factory
   *   The factory for configuration objects.
   * @param \Drupal\Core\Extension\ModuleHandlerInterface $module_handler
   *   The module handler service.
   * @param \Drupal\Core\Routing\UrlGeneratorInterface $url_generator
   *   The url generator.
   * @param \Drupal\geocoder\ProviderPluginManager $geocoder_provider_plugin_manager
   *   The geocoder provider plugin manager.
   */
  public function __construct(
    ConfigFactoryInterface $config_factory,
    ModuleHandlerInterface $module_handler,
    UrlGeneratorInterface $url_generator,
    ProviderPluginManager $geocoder_provider_plugin_manager = NULL
  ) {
    parent::__construct($config_factory);
    $this->moduleHandler = $module_handler;
    $this->urlGenerator = $url_generator;
    $this->geocoderProviderPluginManager = $geocoder_provider_plugin_manager;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container) {
    return new static(
      $container->get('config.factory'),
      $container->get('module_handler'),
      $container->get('url_generator'),
      $container->has('plugin.manager.geocoder.provider') ? $container->get('plugin.manager.geocoder.provider') : NULL
    );
  }

  /**
   * {@inheritdoc}
   */
  public function getFormId() {
    return 'cloud_admin_settings';
  }

  /**
   * {@inheritdoc}
   */
  protected function getEditableConfigNames() {
    return ['cloud.settings'];
  }

  /**
   * {@inheritdoc}
   */
  public function buildForm(array $form, FormStateInterface $form_state) {

    $config = $this->config('cloud.settings');

    $form['location_map'] = [
      '#type' => 'details',
      '#title' => $this->t('Location Map'),
      '#open' => TRUE,
    ];

    $form['location_map']['cloud_location_map_json_url'] = [
      '#type' => 'url',
      '#title' => $this->t('Location Map JSON URL'),
      '#default_value' => $config->get('cloud_location_map_json_url'),
      '#description' => $this->t('The URL of the geojson file to draw the location map.'),
    ];

    if ($this->moduleHandler->moduleExists('geocoder')) {
      $form['geocoder'] = [
        '#type' => 'details',
        '#title' => $this->t('Geocoder'),
        '#open' => TRUE,
      ];

      $providers = $this->geocoderProviderPluginManager->getPlugins();
      $options = [];
      foreach ($providers as $key => $values) {
        $options[$key] = $values['name'];
      }

      $form['geocoder']['cloud_location_geocoder_plugin'] = [
        '#type' => 'select',
        '#title' => $this->t('Geocoder Pligin'),
        '#options' => $options,
        '#default_value' => $config->get('cloud_location_geocoder_plugin'),
        '#description' => $this->t('The Geocoder plugin to get latitude and longitude of cloud service provider location. Please confirrm the options / arguments of selected Geocoder plugin are set at <a href=":url">Geocoder configuration</a>.',
          [':url' => $this->urlGenerator->generate('geocoder.settings')]),
      ];
    }

    return parent::buildForm($form, $form_state);
  }

  /**
   * {@inheritdoc}
   */
  public function submitForm(array &$form, FormStateInterface $form_state) {
    $config = $this->configFactory()->getEditable('cloud.settings');
    $config->set('cloud_location_map_json_url', $form_state->getValue('cloud_location_map_json_url'));
    if ($this->moduleHandler->moduleExists('geocoder')) {
      $config->set('cloud_location_geocoder_plugin', $form_state->getValue('cloud_location_geocoder_plugin'));
    }
    $config->save();
    parent::submitForm($form, $form_state);
  }

}
